/*
 *  Copyright (C) 1997-2007 Sean Kelly
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty. In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, in both source and binary form, subject to the following
 *  restrictions:
 *
 *  o  The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  o  Altered source versions must be plainly marked as such, and must not
 *     be misrepresented as being the original software.
 *  o  This notice may not be removed or altered from any source
 *     distribution.
 */

/**
 *
 */
module sql.Connection;


private import sql.Environment;
private import sql.Exception;
private import sql.Statement;
private import win32.sql;
private import win32.sqlext;
private import win32.sqltypes;


////////////////////////////////////////////////////////////////////////////////
// Connection
////////////////////////////////////////////////////////////////////////////////


/**
 *
 */
class Connection
{
    /**
     *
     */
    this( char[] connstr )
    {
        SQLHDBC hdbc    = SQL_NULL_HDBC;
        RETCODE retcode = SQLAllocHandle( SQL_HANDLE_DBC, Environment.handle, &hdbc );
        process_retcode( retcode, SQL_HANDLE_DBC, Environment.handle );
        scope(failure) if( m_conn is null ) SQLFreeHandle( SQL_HANDLE_DBC, hdbc );

        // NOTE: SQL_ATTR_ODBC_CURSORS provides client caching and cursor
        //       facilities for drivers which do not natively support them.
        //       It apprently caches the entire result set locally.  On Win32
        //       this feature requires the presence of odbccr32.dll for ANSI
        //       applications and odbccu32.dll for Unicode applications.  This
        //       option may be removes once block fetching is implemented in
        //       ResultSet.
        SQLSetConnectAttr( hdbc, SQL_ATTR_ODBC_CURSORS, cast(SQLPOINTER) SQL_CUR_USE_ODBC, 0 );
        process_retcode( retcode, SQL_HANDLE_DBC, hdbc );
        retcode = SQLDriverConnect( hdbc,
                                    null,
                                    cast(SQLCHAR*) connstr.ptr,
                                    connstr.length,
                                    null,
                                    0,
                                    null,
                                    SQL_DRIVER_NOPROMPT );
        process_retcode( retcode, SQL_HANDLE_DBC, hdbc );
        scope(failure) if( m_conn is null ) SQLDisconnect( hdbc );

        m_conn    = new ConnHandle( hdbc );
        m_connstr = connstr;
    }

    this( char[] dsn, char[] user, char[] password )
    {
        SQLHDBC hdbc    = SQL_NULL_HDBC;
        RETCODE retcode = SQLAllocHandle( SQL_HANDLE_DBC, Environment.handle, &hdbc );
        process_retcode( retcode, SQL_HANDLE_DBC, Environment.handle );
        scope(failure) if( m_conn is null ) SQLFreeHandle( SQL_HANDLE_DBC, hdbc );

        // NOTE: SQL_ATTR_ODBC_CURSORS provides client caching and cursor
        //       facilities for drivers which do not natively support them.
        //       It apprently caches the entire result set locally.  On Win32
        //       this feature requires the presence of odbccr32.dll for ANSI
        //       applications and odbccu32.dll for Unicode applications.  This
        //       option may be removes once block fetching is implemented in
        //       ResultSet.
        SQLSetConnectAttr( hdbc, SQL_ATTR_ODBC_CURSORS, cast(SQLPOINTER) SQL_CUR_USE_ODBC, 0 );
        process_retcode( retcode, SQL_HANDLE_DBC, hdbc );
        retcode = SQLConnect( hdbc,
                              cast(SQLCHAR*) dsn.ptr, dsn.length,
                              cast(SQLCHAR*) user.ptr, user.length,
                              cast(SQLCHAR*) password.ptr, password.length );
        process_retcode( retcode, SQL_HANDLE_DBC, hdbc );
        scope(failure) if( m_conn is null ) SQLDisconnect( hdbc );

        m_conn     = new ConnHandle( hdbc );
        m_dsn      = dsn;
        m_user     = user;
        m_password = password;
    }


    ~this()
    {
        m_conn.release();
    }


    /**
     * cursor_type:
     *
     *   SQL_CURSOR_FORWARD_ONLY    = The cursor only scrolls forward (DEFAULT).
     *   SQL_CURSOR_STATIC          = The data in the result set is static.
     *   SQL_CURSOR_KEYSET_DRIVEN   = The driver saves and uses the keys
     *                                for the number of rows specified in
     *                                the SQL_ATTR_KEYSET_SIZE Statement attribute.
     *   SQL_CURSOR_DYNAMIC         = The driver saves and uses only the keys for the rows
     *                                in the rowset.
     *
     * concurrency:
     *
     *   SQL_CONCUR_READ_ONLY       = Cursor is read-only. No updates are allowed (DEFAULT).
     *   SQL_CONCUR_LOCK            = Cursor uses the lowest level of locking sufficient
     *                                to ensure that the row can be updated.
     *   SQL_CONCUR_ROWVER          = Cursor uses optimistic concurrency control,
     *                                comparing row versions such as SQLBase ROWID or Sybase TIMESTAMP.
     *   SQL_CONCUR_VALUES          = Cursor uses optimistic concurrency control, comparing values.
     */
    final Statement prepare( char[] query,
                             SQLUINTEGER cursor_type = SQL_CURSOR_FORWARD_ONLY,
                             SQLUINTEGER concurrency = SQL_CONCUR_READ_ONLY )
    {
        return new Statement( m_conn, query, cursor_type, concurrency );
    }

    final char[] connstr()
    {
        return m_connstr;
    }

    final char[] dsn()
    {
        return m_dsn;
    }

    final char[] user()
    {
        return m_user;
    }

    final char[] password()
    {
        return m_password;
    }

private:
    ConnHandle  m_conn;
    char[]      m_connstr;
    char[]      m_dsn;
    char[]      m_user;
    char[]      m_password;
}
