/*
 *  Copyright (C) 1997-2007 Sean Kelly
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty. In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, in both source and binary form, subject to the following
 *  restrictions:
 *
 *  o  The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  o  Altered source versions must be plainly marked as such, and must not
 *     be misrepresented as being the original software.
 *  o  This notice may not be removed or altered from any source
 *     distribution.
 */

/**
 *
 */
module sql.Environment;


// TODO: Double check use of .length and .size properties to make sure the
//       correct ones are being used.  (I think this has been fixed).
//
// TODO: Eliminate min, declared in Statement.
//
// TODO: Consider doing something about BOOL.


private import sql.Exception;
private import tango.stdc.stdlib : malloc, free;
private import win32.sql;
private import win32.sqlext;
private import win32.sqltypes;


static assert( SQLCHAR.sizeof == char.sizeof, "SQLCHAR and char sizes don't match." );


////////////////////////////////////////////////////////////////////////////////
// Environment
////////////////////////////////////////////////////////////////////////////////


class Environment
{
    static SQLHENV handle()
    {
        return sm_handle;
    }

    static this()
    {
        SQLUINTEGER odbcVersion = SQL_OV_ODBC3;

        RETCODE retcode = SQLAllocHandle( SQL_HANDLE_ENV, null, &sm_handle );
        process_retcode( retcode, SQL_HANDLE_ENV, null );
        scope(failure) cleanup();
        retcode = SQLSetEnvAttr( sm_handle, SQL_ATTR_ODBC_VERSION, cast(SQLPOINTER) odbcVersion, 0 );
        process_retcode( retcode, SQL_HANDLE_ENV, sm_handle );
    }

    static ~this()
    {
        cleanup();
    }

private:
    static void cleanup()
    {
        SQLFreeHandle( SQL_HANDLE_ENV, sm_handle );
        sm_handle = SQL_NULL_HENV;
    }

    static SQLHENV sm_handle = SQL_NULL_HENV;
}


////////////////////////////////////////////////////////////////////////////////
// Handle Manager
////////////////////////////////////////////////////////////////////////////////


private class Managed( Handle, alias cleanup )
{
    final Handle handle()
    {
        return m_handle;
    }

    final void acquire()
    {
        synchronized
        {
            ++m_count;
        }
    }

    final void release()
    {
        synchronized
        {
            if( --m_count > 0 )
                return;
        }
        delete this;
    }

    final int count()
    {
        return m_count;
    }

    this( Handle h )
    {
        m_handle = h;
        m_count  = 1;
    }

    ~this()
    {
        cleanup( m_handle );
    }

    new( size_t size )
    {
        return malloc( size );
    }

    delete( void* ptr )
    {
        free( ptr );
    }

private:
    Handle  m_handle;
    int     m_count;
}


////////////////////////////////////////////////////////////////////////////////
// Handles
////////////////////////////////////////////////////////////////////////////////


/**
 *
 */
private void cleanupHDBC( inout SQLHDBC handle )
{
    if( handle != SQL_NULL_HDBC )
    {
        SQLDisconnect( handle );
        SQLFreeHandle( SQL_HANDLE_DBC, handle );
        handle = SQL_NULL_HDBC;
    }
}


/**
 *
 */
private void cleanupHSTMT( inout SQLHSTMT handle )
{
    if( handle != SQL_NULL_HSTMT )
    {
        SQLFreeStmt( handle, SQL_RESET_PARAMS );
        SQLFreeHandle( SQL_HANDLE_STMT, handle );
        handle = SQL_NULL_HSTMT;
    }
}

alias Managed!(SQLHDBC, cleanupHDBC)   ConnHandle;
alias Managed!(SQLHSTMT, cleanupHSTMT) StmtHandle;


////////////////////////////////////////////////////////////////////////////////
// Null Value
////////////////////////////////////////////////////////////////////////////////


/**
 *
 */
struct NullValue
{

}


NullValue SQLNull; ///


////////////////////////////////////////////////////////////////////////////////
// Utility Functions
////////////////////////////////////////////////////////////////////////////////


/**
 *
 */
package bool isCharType( SQLSMALLINT sqlType )
{
    return sqlType == SQL_C_CHAR;
}


/**
 *
 */
package SQLSMALLINT defaultCType( SQLSMALLINT sqlType, SQLSMALLINT sqlSigned )
{
    switch( sqlType )
    {
        case SQL_CHAR:
        case SQL_VARCHAR:
        case SQL_LONGVARCHAR:
        case SQL_DECIMAL:
        case SQL_NUMERIC:
        case SQL_GUID:
            return SQL_C_CHAR;
        case SQL_WCHAR:
        case SQL_WVARCHAR:
        case SQL_WLONGVARCHAR:
            return SQL_C_WCHAR;
        case SQL_BIT:
            return SQL_C_BIT;
        case SQL_TINYINT:
            return sqlSigned ? SQL_C_STINYINT : SQL_C_UTINYINT;
        case SQL_SMALLINT:
            return sqlSigned ? SQL_C_SSHORT   : SQL_C_USHORT;
        case SQL_INTEGER:
            return sqlSigned ? SQL_C_SLONG    : SQL_C_ULONG;
        case SQL_BIGINT:
            return sqlSigned ? SQL_C_SBIGINT  : SQL_C_UBIGINT;
        case SQL_REAL:
            return SQL_C_FLOAT;
        case SQL_FLOAT:
        case SQL_DOUBLE:
            return SQL_C_DOUBLE;
        case SQL_BINARY:
        case SQL_VARBINARY:
        case SQL_LONGVARBINARY:
            return SQL_C_BINARY;
        case SQL_TYPE_DATE:
            return SQL_C_TYPE_DATE;
        case SQL_TYPE_TIME:
            return SQL_C_TYPE_TIME;
        case SQL_DATETIME:
        case SQL_INTERVAL:
        case SQL_TYPE_TIMESTAMP:
            return SQL_C_TYPE_TIMESTAMP;
        default:
            return sqlType;
    }
}
