/*
 *  Copyright (C) 1997-2007 Sean Kelly
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty. In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, in both source and binary form, subject to the following
 *  restrictions:
 *
 *  o  The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  o  Altered source versions must be plainly marked as such, and must not
 *     be misrepresented as being the original software.
 *  o  This notice may not be removed or altered from any source
 *     distribution.
 */

/**
 *
 */
module sql.Exception;


private import win32.sql;
private import win32.sqlext;
private import win32.sqltypes;


////////////////////////////////////////////////////////////////////////////////
// Event Hooks
////////////////////////////////////////////////////////////////////////////////


private
{
    alias void function( SQLWarning ) warningHandlerType;
    alias void function( SQLError ) errorHandlerType;

    warningHandlerType  warningHandler = null;
    errorHandlerType    errorHandler   = null;
}


/**
 *
 */
void setWarningHandler( warningHandlerType h )
{
    warningHandler = h;
}


/**
 *
 */
void setErrorHandler( errorHandlerType h )
{
    errorHandler = h;
}


////////////////////////////////////////////////////////////////////////////////
// SQL Info
////////////////////////////////////////////////////////////////////////////////


/**
 *
 */
struct SQLInfo
{
    static SQLInfo opCall( SQLSMALLINT htype, SQLHANDLE handle )
    {
        SQLInfo temp;
        temp.m_htype  = htype;
        temp.m_handle = handle;
        return temp;
    }

    static struct Record
    {
        SQLCHAR[]   state;
        SQLINTEGER  code;
        SQLCHAR[]   msg;
    }

    int opApply( int delegate( inout Record ) dg )
    {
        SQLINTEGER  errcode = 0;
        SQLSMALLINT msglen  = 0;
        RETCODE     retcode = SQL_SUCCESS;
        SQLSMALLINT recnum  = 1;
        SQLCHAR[6]  state   = void;
        SQLCHAR[SQL_MAX_MESSAGE_LENGTH] msg = void;
        Record      rec = void;
        int         ret = 0;

        rec.state = state[0 .. $];

        while( retcode != SQL_NO_DATA )
        {
            retcode = SQLGetDiagRec( m_htype,
                                     m_handle,
                                     recnum,
                                     state.ptr,
                                     &rec.code,
                                     msg.ptr,
                                     msg.length - 1,
                                     &msglen );
            process_retcode( retcode, m_htype, m_handle );
            if( retcode != SQL_NO_DATA )
            {
                rec.msg = msg[0 .. msglen];
                ret = dg( rec );
                if( ret != 0 )
                    return ret;
            }
            ++recnum;
        }
        return ret;
    }

private:
    SQLSMALLINT m_htype;
    SQLHANDLE   m_handle;
}


/**
 *
 */
struct SQLWarning
{
    static SQLWarning opCall( SQLSMALLINT htype, SQLHANDLE handle )
    {
        SQLWarning temp;
        temp.m_info = SQLInfo( htype, handle );
        return temp;
    }


    alias SQLInfo.Record Record;


    int opApply( int delegate( inout Record ) dg )
    {
        int ret = 0;
        foreach( rec; m_info )
        {
            if( isWarning( rec.state ) )
            {
                ret = dg( rec );
                if( ret != 0 )
                    return ret;
            }
        }
        return ret;
    }


private:
    SQLInfo m_info;
}


/**
 *
 */
struct SQLError
{
    static SQLError opCall( SQLSMALLINT htype, SQLHANDLE handle )
    {
        SQLError temp;
        temp.m_info = SQLInfo( htype, handle );
        return temp;
    }


    alias SQLInfo.Record Record;


    int opApply( int delegate( inout Record ) dg )
    {
        int ret = 0;
        foreach( rec; m_info )
        {
            if( !isWarning( rec.state ) )
            {
                ret = dg( rec );
                if( ret != 0 )
                    return ret;
            }
        }
        return ret;
    }


private:
    SQLInfo m_info;
}


////////////////////////////////////////////////////////////////////////////////
// SQL Exception
////////////////////////////////////////////////////////////////////////////////


/**
 *
 */
class SQLException :
    public Exception
{
    /**
     *
     */
    static struct Record
    {
        char[]  state;
        int     code;
        char[]  msg;

        static Record opCall( SQLInfo.Record rec )
        {
            // NOTE: If we're coming from a SQLInfo.Record then we know that
            //       the char elements are slices of static buffers, so we must
            //       .dup the data to keep it alive and valid long enough to
            //       report.
            Record temp;
            temp.state = cast(char[]) rec.state.dup;
            temp.code  = rec.code;
            temp.msg   = cast(char[]) rec.msg.dup;
            return temp;
        }
    }


    /**
     *
     */
    this( char[] msg = "General SQL Exception" )
    {
        super( msg );
    }


    /**
     *
     */
    SQLException append( SQLInfo.Record rec )
    {
        m_all ~= Record( rec );
        return this;
    }


    /**
     *
     */
    int opApply( int delegate( inout Record ) dg )
    {
        int ret = 0;
        foreach( rec; m_all )
        {
            ret = dg( rec );
            if( ret != 0 )
                return ret;
        }
        return ret;
    }


private:
    Record[]  m_all;
}


////////////////////////////////////////////////////////////////////////////////
// Event Handling
////////////////////////////////////////////////////////////////////////////////


/**
 *
 */
private bool isWarning( SQLCHAR[] sqlstate )
{
    return sqlstate[0] == '0' && sqlstate[1] == '1';
}


/**
 *
 */
private void process_warnings( SQLSMALLINT htype, SQLHANDLE handle )
{
    if( warningHandler !is null )
    {
        warningHandler( SQLWarning( htype, handle ) );
    }
    else
    {
        // ignore by default
        foreach( rec; SQLWarning( htype, handle ) )
        {
        }
    }
}


/**
 *
 */
private void process_errors( SQLSMALLINT htype, SQLHANDLE handle )
{
    if( errorHandler !is null )
    {
        errorHandler( SQLError( htype, handle ) );
    }
    else
    {
        SQLException e = new SQLException();

        foreach( rec; SQLError( htype, handle ) )
        {
            e.append( rec );
        }
        throw e;
    }
}


/**
 *
 */
package void process_retcode( RETCODE retcode, SQLSMALLINT htype, SQLHANDLE handle )
{
    switch( retcode )
    {
    case SQL_SUCCESS:
    case SQL_NO_DATA:
        break;
    case SQL_SUCCESS_WITH_INFO:
        process_warnings( htype, handle );
        break;
    default:
        // SQL_NEED_DATA
        // SQL_STILL_EXECUTING
        // SQL_ERROR
        // SQL_INVALID_HANDLE
        process_errors( htype, handle );
    }
}
