/*
 *  Copyright (C) 1997-2007 Sean Kelly
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty. In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, in both source and binary form, subject to the following
 *  restrictions:
 *
 *  o  The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  o  Altered source versions must be plainly marked as such, and must not
 *     be misrepresented as being the original software.
 *  o  This notice may not be removed or altered from any source
 *     distribution.
 */

/**
 *
 */
module sql.ResultSet;


private import sql.Environment;
private import sql.Exception;
private import tango.core.Exception;
private import tango.stdc.stdlib : atoi, atof, atoll;
private import win32.sql;
private import win32.sqlext;
private import win32.sqltypes;


////////////////////////////////////////////////////////////////////////////////
// Resultset
////////////////////////////////////////////////////////////////////////////////


/**
 *
 */
class ResultSet
{
    /**
     *
     */
    static class Column
    {
        /**
         *
         */
        final SQLUSMALLINT index()
        {
            return m_index;
        }


        /**
         *
         */
        final char[] name()
        {
            return cast(char[]) m_name;
        }


        /**
         *
         */
        final SQLSMALLINT dataType()
        {
            return m_dataType;
        }


        /**
         *
         */
        final SQLINTEGER size()
        {
            return m_size;
        }


        /**
         *
         */
        final SQLINTEGER length()
        {
            return m_length;
        }


        /**
         *
         */
        final void[] data()
        {
            return m_data[0 .. m_length];
        }


        /**
         *
         */
        final bool isNull()
        {
            return m_length == SQL_NULL_DATA;
        }


        /**
         *
         */
        final bool opEquals( NullValue val )
        {
            return isNull();
        }


        ////////////////////////////////////////////////////////////////////////
        // char/string
        ////////////////////////////////////////////////////////////////////////


        /**
         *
         */
        final char* asUtf8z()
        {
            return isNull() ? null : cast(char*) m_data.ptr;
        }


        /**
         *
         */
        final char[] asUtf8()
        {
            return isNull() ? null : cast(char[]) m_data[0 .. m_length];
        }


        ////////////////////////////////////////////////////////////////////////
        // date/time/timestamp
        ////////////////////////////////////////////////////////////////////////


        /**
         *
         */
        final SQL_DATE_STRUCT asSQLDate()
        {
            if( isNull() )
                return SQL_DATE_STRUCT.init;
            return *cast(SQL_DATE_STRUCT*) m_data.ptr;
        }


        /**
         *
         */
        final SQL_TIME_STRUCT asSQLTime()
        {
            if( isNull() )
                return SQL_TIME_STRUCT.init;
            return *cast(SQL_TIME_STRUCT*) m_data.ptr;
        }


        /**
         *
         */
        final SQL_TIMESTAMP_STRUCT asSQLTimeStamp()
        {
            if( isNull() )
                return SQL_TIMESTAMP_STRUCT.init;
            return *cast(SQL_TIMESTAMP_STRUCT*) m_data.ptr;
        }


        ////////////////////////////////////////////////////////////////////////
        // numeric
        ////////////////////////////////////////////////////////////////////////


        /**
         *
         */
        final bool asBool()
        {
            if( isNull() )
                return bool.init;
            else if( isCharType( m_dataType ) )
                return atoi( asUtf8z() ) != 0;
            else
                return *cast(bool*) m_data.ptr;
        }


        /**
         *
         */
        final double asDouble()
        {
            if( isNull() )
                return double.init;
            else if( isCharType( m_dataType ) )
                return atof( asUtf8z() );
            else
                return *cast(double*) m_data.ptr;
        }


        /**
         *
         */
        final int asInteger()
        {
            if( isNull() )
                return int.init;
            else if( isCharType( m_dataType ) )
                return atoi( asUtf8z() );
            else
                return *cast(int*) m_data.ptr;
        }


    protected:
        //
        //
        //
        final void init( SQLHSTMT hstmt, SQLUSMALLINT n )
        {
            SQLSMALLINT sqlType,
                        sqlSigned,
                        len;
            //
            // column number
            //
            m_index = n;
            //
            // column name
            //
            RETCODE retcode = SQLColAttribute( hstmt, m_index, SQL_DESC_NAME, null, 0, &len, null );
            process_retcode( retcode, SQL_HANDLE_STMT, hstmt );
            m_name.length = ++len;
            retcode = SQLColAttribute( hstmt, m_index, SQL_DESC_NAME, m_name.ptr, len, null, null );
            process_retcode( retcode, SQL_HANDLE_STMT, hstmt );
            //
            // column datatype
            //
            retcode = SQLColAttribute( hstmt, m_index, SQL_DESC_TYPE, null, 0, null, &sqlType );
            process_retcode( retcode, SQL_HANDLE_STMT, hstmt );
            retcode = SQLColAttribute( hstmt, m_index, SQL_DESC_UNSIGNED, null, 0, null, &sqlSigned );
            process_retcode( retcode, SQL_HANDLE_STMT, hstmt );
            m_dataType = defaultCType( sqlType, sqlSigned );
            //
            // column data size
            //
            retcode = SQLColAttribute( hstmt, m_index, SQL_DESC_OCTET_LENGTH, null, 0, null, &m_size );
            process_retcode( retcode, SQL_HANDLE_STMT, hstmt );
            //
            // if char type, bump m_size by 1 for trailing null
            //
            if( isCharType( m_dataType ) )
                ++m_size;
            //
            // initialize data buffer
            //
            m_data.length = m_size;
            //
            // bind column
            //
            retcode = SQLBindCol( hstmt, m_index, m_dataType, m_data.ptr, m_size, &m_length );
            process_retcode( retcode, SQL_HANDLE_STMT, hstmt );
        }


    private:
        SQLUSMALLINT    m_index     = 0;
        SQLCHAR[]       m_name      = null;
        SQLSMALLINT     m_dataType  = 0;
        SQLINTEGER      m_size      = 0;
        SQLINTEGER      m_length    = SQL_NULL_DATA;
        byte[]          m_data      = null;
    }


    /**
     *
     */
    this( ConnHandle conn, StmtHandle stmt )
    {
        m_conn = conn;
        m_stmt = stmt;

        init_rs();

        m_conn.acquire();
        m_stmt.acquire();
    }


    /**
     *
     */
    ~this()
    {
        SQLFreeStmt( m_stmt.handle, SQL_CLOSE );
        SQLFreeStmt( m_stmt.handle, SQL_UNBIND );

        m_stmt.release();
        m_conn.release();
    }


    /**
     *
     */
    final size_t numCols()
    {
        return m_cols.length;
    }


    /**
     *
     */
    final bool nextResultSet()
    {
        RETCODE retcode = SQLMoreResults( m_stmt.handle );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
        init_rs();
        return retcode != SQL_NO_DATA;
    }


    /**
     *
     */
    final bool next()
    {
        RETCODE retcode = SQLFetchScroll( m_stmt.handle, SQL_FETCH_NEXT, 0 );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
        return retcode != SQL_NO_DATA;
    }


    /**
     *
     */
    final bool prev()
    {
        RETCODE retcode = SQLFetchScroll( m_stmt.handle, SQL_FETCH_PREV, 0 );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
        return retcode != SQL_NO_DATA;
    }


    /**
     *
     */
    final bool first()
    {
        RETCODE retcode = SQLFetchScroll( m_stmt.handle, SQL_FETCH_FIRST, 0 );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
        return retcode != SQL_NO_DATA;
    }


    /**
     *
     */
    final bool last()
    {
        RETCODE retcode = SQLFetchScroll( m_stmt.handle, SQL_FETCH_LAST, 0 );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
        return retcode != SQL_NO_DATA;
    }


    /**
     *
     */
    final void update()
    {
        RETCODE retcode = SQLSetPos( m_stmt.handle, 0, SQL_UPDATE, SQL_LOCK_NO_CHANGE );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
    }


    /**
     *
     */
    final Column opIndex( size_t n )
    {
        // NOTE: This should probably be moved into an "in" block, or the
        //       built-in range checking on non-release builds could be
        //       relied upon instead.
        if( n >= m_cols.length )
            throw new NoSuchElementException( "Index Out of Bounds" );
        return m_cols[n];
    }


private:
    //
    //
    //
    void init_rs()
    {
        SQLSMALLINT numcols;
        RETCODE retcode = SQLNumResultCols( m_stmt.handle, &numcols );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
        if( numcols > 0 )
        {
            m_cols.length = numcols;
            foreach( i, inout col; m_cols )
            {
                if( col is null )
                    col = new Column();
                col.init( m_stmt.handle, i + 1 );
            }
        }
    }


    ConnHandle  m_conn;
    StmtHandle  m_stmt;
    Column[]    m_cols;
}
