/*
 *  Copyright (C) 1997-2007 Sean Kelly
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty. In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, in both source and binary form, subject to the following
 *  restrictions:
 *
 *  o  The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  o  Altered source versions must be plainly marked as such, and must not
 *     be misrepresented as being the original software.
 *  o  This notice may not be removed or altered from any source
 *     distribution.
 */

/**
 *
 */
module sql.Statement;


private import sql.Environment;
private import sql.Exception;
private import sql.ResultSet;
private import tango.core.Exception;
private import tango.stdc.string : memcpy;
private import tango.sys.win32.Types : BOOL, TRUE, FALSE;
private import win32.sql;
private import win32.sqlext;
private import win32.sqltypes;


private T min(T)( T a, T b ) { return a < b ? a : b; }


////////////////////////////////////////////////////////////////////////////////
// Statement
////////////////////////////////////////////////////////////////////////////////


/**
 *
 */
class Statement
{
    /**
     *
     */
    static class Param
    {
        /**
         *
         */
        this( StmtHandle stmt )
        {
            m_stmt = stmt;
            // NOTE: Since the lifetime of this object is bound to Statement,
            //       there is no reason to perform a m_stmt.acquire() here.
            //       The reference to StmtHandle is merely for convenience.
        }


        /**
         *
         */
        final void bindTo( SQLUSMALLINT index )
        {
            m_index = index;
            RETCODE retcode = SQLDescribeParam( m_stmt.handle, m_index,
                                                &m_dataType,   &m_size,
                                                &m_digits,     &m_isNullable );
            process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
            setDataType( m_dataType );
            m_data.length = m_size;
            bind();
        }


        /**
         *
         */
        final SQLUSMALLINT index()
        {
            return m_index;
        }


        /**
         *
         */
        final SQLSMALLINT dataType()
        {
            return m_dataType;
        }


        /**
         *
         */
        final SQLUINTEGER size()
        {
            return m_size;
        }


        /**
         *
         */
        final SQLSMALLINT digits()
        {
            return m_digits;
        }


        /**
         *
         */
        final SQLSMALLINT isNullable()
        {
            return m_isNullable;
        }


        /**
         *
         */
        final SQLSMALLINT ioType()
        {
            return m_ioType;
        }


        /**
         *
         */
        final SQLINTEGER length()
        {
            return m_length;
        }


        /**
         *
         */
        final void[] data()
        {
            return cast(void[]) m_data[0 .. m_length];
        }


        /**
         *
         */
        final bool isBound()
        {
            return m_data != null;
        }


        /**
         *
         */
        final bool isNull()
        {
            return m_isNull;
        }


        /**
         *
         */
        final Param opAssign( NullValue val )
        {
            if( isNull() )
                return this;
            m_isNull = true;
            m_length = SQL_NULL_DATA;
            bind();
            return this;
        }


        /**
         *
         */
        final Param opAssign( bool param )
        {
            // NOTE: This test should perhaps go in an "in" clause.  Same for
            //       the other opAssigns.
            if( dataType() != SQL_BIT )
                throw new Exception( "Parameter is already bound as a different datatype" );
            // TODO: This should maybe be SQLBIT or something.
            BOOL temp = ( param ? TRUE : FALSE );
            setData( &temp, temp.sizeof );
            return this;
        }


        /**
         *
         */
        final Param opAssign( int param )
        {
            if( dataType() != SQL_INTEGER )
                throw new Exception( "Parameter is already bound as a sifferent datatype" );
            setData( &param, param.sizeof );
            return this;
        }


        /**
         *
         */
        final Param opAssign( double param )
        {
            if( dataType() != SQL_DOUBLE )
                throw new Exception( "Parameter is already bound as a different datatype" );
            setData( &param, param.sizeof );
            return this;
        }


        /**
         *
         */
        final Param opAssign( char[] param )
        {
            if( dataType() != SQL_CHAR &&
                dataType() != SQL_VARCHAR )
            {
                throw new Exception( "Parameter is already bound as a different datatype" );
            }
            setData( param.ptr, param.length );
            return this;
        }

    private:
        //
        //
        //
        void setDataType( SQLSMALLINT dataType )
        {
            const STR_MAX = 255;

            m_dataType = dataType;
            switch( m_dataType )
            {
            case SQL_BIT:
                m_size      = BOOL.sizeof;
                m_ioType    = SQL_PARAM_INPUT;
                m_length    = BOOL.sizeof;
                m_bindAs    = SQL_C_BIT;
                break;
            case SQL_INTEGER:
                m_size      = SQLINTEGER.sizeof;
                m_ioType    = SQL_PARAM_INPUT;
                m_length    = SQLINTEGER.sizeof;
                m_bindAs    = SQL_C_SLONG;
                break;
            case SQL_DOUBLE:
                m_size      = SQLDOUBLE.sizeof;
                m_ioType    = SQL_PARAM_INPUT;
                m_length    = SQLDOUBLE.sizeof;
                m_bindAs    = SQL_C_DOUBLE;
                break;
            case SQL_CHAR:
            case SQL_VARCHAR:
                m_size      = ( m_size == 0 ? STR_MAX * SQLCHAR.sizeof : m_size );
                m_ioType    = SQL_PARAM_INPUT;
                m_length    = SQL_NTS;
                m_bindAs    = SQL_C_CHAR;
                break;
            default:
                throw new Exception( "Paramater datatype not recognized for bind" );
            }
        }


        //
        //
        //
        void setData( void* x, size_t len )
        {
            if( isNull() )
            {
                m_isNull = false;
                setDataType( m_dataType );
                bind();
            }
            if( m_dataType == SQL_CHAR ||
                m_dataType == SQL_VARCHAR )
            {
                size_t cpylen = min!(size_t)( len + 1, m_size );
                memcpy( m_data.ptr, x, cpylen );
                m_data[cpylen - 1] = 0;
            }
            else
            {
                memcpy( m_data.ptr, x, min!(size_t)( len, m_size ) );
            }
        }


        //
        //
        //
        void bind()
        {
            RETCODE retcode = SQLBindParameter( m_stmt.handle, m_index,
                                                m_ioType,      m_bindAs,
                                                m_dataType,    m_size,
                                                m_digits,      m_data.ptr,
                                                m_size,        &m_length );
            process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
        }


        StmtHandle      m_stmt       = null;
        SQLUSMALLINT    m_index      = 0;
        SQLSMALLINT     m_dataType   = SQL_CHAR;
        SQLUINTEGER     m_size       = 0;
        SQLSMALLINT     m_digits     = 0;
        SQLSMALLINT     m_isNullable = SQL_NO_NULLS;
        SQLSMALLINT     m_ioType     = SQL_PARAM_INPUT;
        SQLINTEGER      m_length     = SQL_NTS;
        SQLSMALLINT     m_bindAs     = 0;
        byte[]          m_data       = null;
        bool            m_isNull     = false;
    }


    /**
     *
     */
    this( ConnHandle conn,
          char[] query,
          SQLUINTEGER cursorType,
          SQLUINTEGER concurrency )
    {
        m_conn  = conn;
        m_query = query;

        SQLHSTMT hstmt   = SQL_NULL_HSTMT;
        RETCODE  retcode = SQLAllocHandle( SQL_HANDLE_STMT, m_conn.handle, &hstmt );
        process_retcode( retcode, SQL_HANDLE_STMT, m_conn.handle );
        scope(failure) if( m_stmt is null ) SQLFreeHandle( SQL_HANDLE_STMT, hstmt );

        retcode = SQLSetStmtAttr( hstmt, SQL_ATTR_CURSOR_TYPE, cast(SQLPOINTER) cursorType, 0 );
        process_retcode( retcode, SQL_HANDLE_STMT, hstmt );
        retcode = SQLSetStmtAttr( hstmt, SQL_ATTR_CONCURRENCY, cast(SQLPOINTER) concurrency, 0 );
        process_retcode( retcode, SQL_HANDLE_STMT, hstmt );
        retcode = SQLSetStmtAttr( hstmt, SQL_ATTR_CURSOR_SCROLLABLE, cast(SQLPOINTER) SQL_SCROLLABLE, 0 );
        process_retcode( retcode, SQL_HANDLE_STMT, hstmt );
        retcode = SQLPrepare( hstmt, cast(SQLCHAR*) m_query.ptr, m_query.length );
        process_retcode( retcode, SQL_HANDLE_STMT, hstmt );
        m_stmt = new StmtHandle( hstmt );
        scope(failure) m_stmt.release();

        SQLSMALLINT numparams;
        retcode = SQLNumParams( m_stmt.handle, &numparams );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
        if( numparams > 0 )
        {
            m_params.length = numparams;

            foreach( i, inout param; m_params )
            {
                param = new Param( m_stmt );
                param.bindTo( i + 1 );
            }
        }

        m_conn.acquire();
    }


    ~this()
    {
        m_stmt.release();
        m_conn.release();
    }


    /**
     *
     */
    final void execute()
    {
        RETCODE retcode = SQLExecute( m_stmt.handle );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
        retcode = SQLFreeStmt( m_stmt.handle, SQL_CLOSE );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
    }


    /**
     *
     */
    final ResultSet open()
    {
        // NOTE: Both of these conditions may be better placed in an "in" clause.
        if( m_stmt.count() > 1 )
        {
            throw new Exception( "Only allowed one open ResultSet per statement" );
        }
        foreach( param; m_params )
        {
            if( !param.isBound() )
            {
                throw new Exception( "Must set all parameters before opening ResultSet" );
            }
        }
        RETCODE retcode = SQLExecute( m_stmt.handle );
        process_retcode( retcode, SQL_HANDLE_STMT, m_stmt.handle );
        return new ResultSet( m_conn, m_stmt );
    }


    /**
     *
     */
    final char[] query()
    {
        return m_query;
    }


    /**
     *
     */
    final Param opIndex( size_t n )
    {
        if( n >= m_params.length )
            throw new NoSuchElementException( "Index out of bounds" );
        return m_params[n];
    }


    /**
     *
     */
    final Param opIndex( int n )
    {
        return opIndex( cast(size_t) n );
    }


    /**
     *
     */
    final Param opIndexAssign( NullValue param, size_t n )
    {
        if( n >= m_params.length )
            throw new NoSuchElementException( "Index out of bounds" );
        return m_params[n] = param;
    }


    /**
     *
     */
    final Param opIndexAssign( NullValue param, int n )
    {
        return opIndexAssign( param, cast(size_t) n );
    }


    /**
     *
     */
    final Param opIndexAssign( bool param, size_t n )
    {
        if( n >= m_params.length )
            throw new NoSuchElementException( "Index out of bounds" );
        return m_params[n] = param;
    }


    /**
     *
     */
    final Param opIndexAssign( bool param, int n )
    {
        return opIndexAssign( param, cast(size_t) n );
    }


    /**
     *
     */
    final Param opIndexAssign( int param, size_t n )
    {
        if( n >= m_params.length )
            throw new NoSuchElementException( "Index but of bounds" );
        return m_params[n] = param;
    }


    /**
     *
     */
    final Param opIndexAssign( int param, int n )
    {
        return opIndexAssign( param, cast(size_t) n );
    }


    /**
     *
     */
    final Param opIndexAssign( double param, size_t n )
    {
        if( n >= m_params.length )
            throw new NoSuchElementException( "Index out of bounds" );
        return m_params[n] = param;
    }


    /**
     *
     */
    final Param opIndexAssign( double param, int n )
    {
        return opIndexAssign( param, cast(size_t) n );
    }

    /**
     *
     */
    final Param opIndexAssign( char[] param, size_t n )
    {
        if( n >= m_params.length )
            throw new NoSuchElementException( "Index out of bounds" );
        return m_params[n] = param;
    }


    /**
     *
     */
    final Param opIndexAssign( char[] param, int n )
    {
        return opIndexAssign( param, cast(size_t) n );
    }


private:
    ConnHandle  m_conn;
    StmtHandle  m_stmt;
    char[]      m_query;
    Param[]     m_params;
}
